#
# [name] ion.const.py
#
# Written by Yoshikazu NAKAJIMA
#

#-- STR

STR_ANY = '_ANY'
STR_NULL = ''
STR_TIME_PRESENT = '_PRESENT'
STR_TIME_NOW = STR_TIME_PRESENT
STR_TIME_LATEST = '_LATEST'

ANYSTR = STR_ANY
NULLSTR = STR_NULL
TIME_PRESENT = STR_TIME_PRESENT
TIME_NOW = STR_TIME_NOW
TIME_LATEST = STR_TIME_LATEST

#-- KEYS

__KEY_UNKNOWN = 'unknown'
__KEY_SEMANTICS = 'semantics'
__KEY_ENTITY = 'entity'
__KEY_BASEENTITY = 'baseentity'
__KEY_ROLE = 'role'

KEY_UNKNOWN = __KEY_UNKNOWN
KEY_SEMANTICS = __KEY_SEMANTICS
KEY_ROLE = __KEY_ROLE
KEY_ENTITY = __KEY_ENTITY
KEY_BASEENTITY = __KEY_BASEENTITY

#-- SLOT contents

__SLOT_NULL = STR_NULL
__SLOT_ANY = STR_ANY

SLOT_NULL = __SLOT_NULL
SLOT_ANY = __SLOT_ANY

NULL_SLOT = SLOT_NULL
ANY_SLOT = SLOT_ANY

__KEY_INDEX = 'index'
__SLOT_INDEX = __KEY_INDEX
KEY_INDEX = __KEY_INDEX
SLOT_INDEX = __SLOT_INDEX

#-- SLOT KEYS

# spatial identifier

__KEY_SPATIAL = 'spatial'
__KEY_DIMENSION = 'dimension'    # 次元
__KEY_COORDINATE = 'coordinate'  # 座標

__DEFAULT_DIMENSION = 3

KEY_SPATIAL = __KEY_SPATIAL
KEY_SI = KEY_SPATIAL
KEY_DIMENSION = __KEY_DIMENSION
KEY_COORDINATE = __KEY_COORDINATE

DEFAULT_DIMENSION = __DEFAULT_DIMENSION

# temporal identifier

__KEY_TEMPORAL = 'temporal'
__KEY_TIME = 'time'
__KEY_TIMEPERIOD = 'timeperiod'

KEY_TEMPORAL = __KEY_TEMPORAL
KEY_TI = KEY_TEMPORAL
KEY_TIME = __KEY_TIME
KEY_TIMEPERIOD = __KEY_TIMEPERIOD

# optional identifier

__KEY_OPTIONAL = 'optional'

KEY_OPTIONAL = __KEY_OPTIONAL
KEY_OI = KEY_OPTIONAL

# data body 

__KEY_DATABODY = 'databody'

KEY_DATABODY = __KEY_DATABODY

# data property

__KEY_DATAPROPERTY = 'dataproperty'
__KEY_DATAFORMAT = 'format'
__KEY_DATAUNIT = 'unit'

KEY_DATAPROPERTY = __KEY_DATAPROPERTY
KEY_DATAFORMAT = __KEY_DATAFORMAT
KEY_DATAUNIT = __KEY_DATAUNIT

#-- iON constants

__NAMESPACE_TMDUBMI = 'https://www.tmd.ac.jp/bmi/'

NAMESPACE_TMDUBMI = __NAMESPACE_TMDUBMI

__DEFAULT_NAMESPACE = NAMESPACE_TMDUBMI
__DEFAULT_SLOT = None 
__DEFAULT_STRSLOT = NULL_SLOT

DEFAULT_NAMESPACE = __DEFAULT_NAMESPACE
DEFAULT_SLOT = __DEFAULT_SLOT
DEFAULT_STRSLOT = __DEFAULT_STRSLOT

# others

__KEY_FILENAME = 'filename'
__TIMEDESCRIPTION = '%Y-%m-%d %H:%M:%S.%f'

KEY_FILENAME = __KEY_FILENAME
TIMEDESCRIPTION = __TIMEDESCRIPTION

#-- main

if __name__ == '__main__':
	pass
